--------------------------------------------------------------------------------
-- Part 2: The Simply-Typed Lambda Calculus with ⊤                            --
--------------------------------------------------------------------------------

module Exam.Part2.Specification where

-- Imports ---------------------------------------------------------------------

open import Data.Nat using (ℕ; zero; suc)
open import Relation.Binary.PropositionalEquality using (_≡_)

open import Exam.Util.Fin using (Fin; zero; suc)

-- Operators -------------------------------------------------------------------

infix  9 `_
infixr 7 _`→_ 
infixl 7 _·_
infix  6 `λ_
infixl 5 _,_
infix  4 _–→_
infix  4 _⊢_⦂_

-- Syntax ----------------------------------------------------------------------

data Type : Set where
  -- Function Type
  _`→_  : Type → Type → Type

  -- Unit type
  `⊤    : Type

-- A `Term n` is a term which is allowed to use `n` different free variables.
-- It can use each variable as often as it wants (not using it is also okay).
data Term (n : ℕ) : Set where
  -- Variable
  --
  -- We represent variables as DeBruijn-Indices, i.e. a variable of a
  -- term with n different free variables is a number between 0 and n-1.
  `_  : Fin n → Term n

  -- Lambda Function
  --
  -- If a lambda term is allowed to use `n` different free variables,
  -- then the body is allowed to additionally use the variable bound
  -- by the lambda. Inside of the body, this variable is 0, and a
  -- variable `x` from outside of the body is referred to as variable
  -- `suc x` inside of the body (Debruijn-indices).
  `λ_ : Term (suc n) → Term n

  -- Function application
  _·_ : Term n → Term n → Term n

  -- Unit term
  `tt : Term n

-- Typing Contexts -------------------------------------------------------------

-- A typing context `Context n` is a list of exactly `n` types.
data Context : ℕ → Set where
  ∅   : Context zero
  _,_ : ∀ {n} → Context n → Type → Context (suc n)

-- `lookup Γ x` retrieves the type for variable `x`, i.e. the type at index `x`.
lookup : ∀ {n} → Context n → Fin n → Type
lookup (Γ , t) zero    = t
lookup (Γ , t) (suc x) = lookup Γ x

-- Typing ----------------------------------------------------------------------

data _⊢_⦂_ {n} (Γ : Context n) : Term n → Type → Set where

  ⊢-` : ∀ {t} x
    → lookup Γ x ≡ t
      --------------
    → Γ ⊢ ` x ⦂ t

  ⊢-λ : ∀ {e t₁ t₂}
    → Γ , t₁ ⊢ e ⦂ t₂
      -------------------
    → Γ ⊢ `λ e ⦂ t₁ `→ t₂

  ⊢-· : ∀ {e₁ e₂ t₁ t₂}
    → Γ ⊢ e₁ ⦂ t₁ `→ t₂
    → Γ ⊢ e₂ ⦂ t₁ 
      -----------------
    → Γ ⊢ e₁ · e₂ ⦂ t₂

  ⊢-tt :
      ------------
      Γ ⊢ `tt ⦂ `⊤

-- Semantics -------------------------------------------------------------------

-- Renamings

-- A renaming `Ren m n` is a function that maps a number between 0 and m-1
-- to a number between 0 and n-1. The numbers represent the different
-- variables of a term as DeBrujn-indices.
Ren : ℕ → ℕ → Set
Ren m n = Fin m → Fin n

-- Extending a renaming such that it still works when we apply
-- it to the body of a lambda term, where there is one more
-- variable, which should remain unchanged, i.e. which should be
-- replaced by itself.
extᵣ : ∀ {m n} → Ren m n → Ren (suc m) (suc n)
extᵣ ρ zero    = zero
extᵣ ρ (suc x) = suc (ρ x)

-- Applying a renaming to a term, i.e. replacing each of the `m`
-- different variables of the term with one of `n` different variables
-- according to the renaming, which results in a term with `n`
-- different variables.
ren : ∀ {m n}
  → Ren m n
  → Term m
  → Term n
ren ρ (` x)     = ` (ρ x)
ren ρ (`λ e)    = `λ (ren (extᵣ ρ) e)
ren ρ (e₁ · e₂) = (ren ρ e₁) · (ren ρ e₂)
ren ρ `tt       = `tt

-- The weakening renaming replaces each variable with its successor.
-- This is useful if we want to substitute a term from the context of a
-- lambda term into the body of a lambda term.
wk : ∀ {n} → Ren n (suc n)
wk x = suc x

-- Substitutions

-- A substitution `Sub m n` is a function that maps each of the `m`
-- variables to a Term with `n` variables.
Sub : ℕ → ℕ → Set
Sub m n = Fin m → Term n

-- Like extᵣ but for substitutions.
extₛ : ∀ {m n} → Sub m n → Sub (suc m) (suc n)
extₛ σ zero    = ` zero
extₛ σ (suc x) = ren wk (σ x)

-- Applying a substitution to a term, i.e. replacing each of the `m`
-- different variables of the term with a term with `n` different
-- variables, which produces a term with `n` different variables.
sub : ∀ {m n}
  → Sub m n
  → Term m
  → Term n
sub σ (` x)     = σ x
sub σ (`λ e)    = `λ (sub (extₛ σ) e)
sub σ (e₁ · e₂) = (sub σ e₁) · (sub σ e₂)
sub σ `tt       = `tt

-- The singleton substitution, which replaces variable 0 with e,
-- and decrements all other variables.
0↦ : ∀ {n} → Term n → Sub (suc n) n
(0↦ e) zero    = e
(0↦ e) (suc x) = ` x

-- Nicer syntax for applying a singleton substitution to a term.
_[_] : ∀ {n} → Term (suc n) → Term n → Term n
e₁ [ e₂ ] = sub (0↦ e₂) e₁

-- Values

data Value : Term 0 → Set where
  val-λ : ∀ {e : Term 1}
      ------------
    → Value (`λ e)

  val-tt :
      ---------
      Value `tt

-- Reduction

data _–→_ : Term 0 → Term 0 → Set where
  β-λ : ∀ {e₁ : Term 1} {e₂ : Term 0}
    → Value e₂
    -------------------------
    → (`λ e₁) · e₂ –→ e₁ [ e₂ ]

  ξ-·₁ : {e₁ e₂ e₁' : Term 0}
    → e₁ –→ e₁'
    -------------------
    → e₁ · e₂ –→ e₁' · e₂ 

  ξ-·₂ : {e₁ e₂ e₂' : Term 0}
    → Value e₁
    → e₂ –→ e₂'
    -------------------
    → e₁ · e₂ –→ e₁ · e₂'
